@@ -1,26 +1,30 @@
-use crate::panes::AnsiCode;
+use crate::output::CharacterChunk;
 use crate::panes::{
     grid::Grid,
-    terminal_character::{
-        CharacterStyles, CursorShape, TerminalCharacter, EMPTY_TERMINAL_CHARACTER,
-    },
+    terminal_character::{CursorShape, TerminalCharacter, EMPTY_TERMINAL_CHARACTER},
 };
+use crate::panes::{AnsiCode, LinkHandler};
 use crate::pty::VteBytes;
 use crate::tab::Pane;
+use crate::ClientId;
+use std::cell::RefCell;
+use std::collections::{HashMap, HashSet};
 use std::fmt::Debug;
 use std::os::unix::io::RawFd;
+use std::rc::Rc;
 use std::time::{self, Instant};
 use zellij_utils::pane_size::Offset;
 use zellij_utils::{
     pane_size::{Dimension, PaneGeom},
     position::Position,
+    shared::make_terminal_title,
     vte,
-    zellij_tile::data::{Palette, PaletteColor},
+    zellij_tile::data::{InputMode, Palette, PaletteColor},
 };
 
 pub const SELECTION_SCROLL_INTERVAL_MS: u64 = 10;
 
-use crate::ui::pane_boundaries_frame::PaneFrame;
+use crate::ui::pane_boundaries_frame::{FrameParams, PaneFrame};
 
 #[derive(PartialEq, Eq, Ord, PartialOrd, Hash, Clone, Copy, Debug)]
 pub enum PaneId {
@@ -42,9 +46,10 @@ pub struct TerminalPane {
     selection_scrolled_at: time::Instant,
     content_offset: Offset,
     pane_title: String,
-    frame: Option<PaneFrame>,
-    frame_color: Option<PaletteColor>,
+    pane_name: String,
+    frame: HashMap<ClientId, PaneFrame>,
     borderless: bool,
+    fake_cursor_locations: HashSet<(usize, usize)>, // (x, y) - these hold a record of previous fake cursors which we need to clear on render
 }
 
 impl Pane for TerminalPane {
@@ -91,8 +96,8 @@ impl Pane for TerminalPane {
         self.reflow_lines();
     }
     fn handle_pty_bytes(&mut self, bytes: VteBytes) {
-        for byte in bytes.iter() {
-            self.vte_parser.advance(&mut self.grid, *byte);
+        for &byte in &bytes {
+            self.vte_parser.advance(&mut self.grid, byte);
         }
         self.set_should_render(true);
     }
@@ -133,6 +138,19 @@ impl Pane for TerminalPane {
                     return "OA".as_bytes().to_vec();
                 }
             }
+
+            [27, 91, 72] => {
+                // home key
+                if self.grid.cursor_key_mode {
+                    return vec![27, 79, 72]; // ESC O H
+                }
+            }
+            [27, 91, 70] => {
+                // end key
+                if self.grid.cursor_key_mode {
+                    return vec![27, 79, 70]; // ESC O F
+                }
+            }
             [27, 91, 66] => {
                 // down arrow
                 if self.grid.cursor_key_mode {
@@ -172,9 +190,7 @@ impl Pane for TerminalPane {
     fn render_full_viewport(&mut self) {
         // this marks the pane for a full re-render, rather than just rendering the
         // diff as it usually does with the OutputBuffer
-        if self.frame.is_some() {
-            self.frame.replace(PaneFrame::default());
-        }
+        self.frame.clear();
         self.grid.render_full_viewport();
     }
     fn selectable(&self) -> bool {
@@ -183,88 +199,148 @@ impl Pane for TerminalPane {
     fn set_selectable(&mut self, selectable: bool) {
         self.selectable = selectable;
     }
-    fn render(&mut self) -> Option<String> {
+    fn render(
+        &mut self,
+        _client_id: Option<ClientId>,
+    ) -> Option<(Vec<CharacterChunk>, Option<String>)> {
         if self.should_render() {
-            let mut vte_output = String::new();
-            let mut character_styles = CharacterStyles::new();
-            if self.grid.clear_viewport_before_rendering {
-                for line_index in 0..self.grid.height {
-                    let x = self.get_content_x();
-                    let y = self.get_content_y();
-                    vte_output.push_str(&format!(
-                        "\u{1b}[{};{}H\u{1b}[m",
-                        y + line_index + 1,
-                        x + 1
-                    )); // goto row/col and reset styles
-                    for _col_index in 0..self.grid.width {
-                        vte_output.push(EMPTY_TERMINAL_CHARACTER.character);
-                    }
+            let mut raw_vte_output = String::new();
+            let content_x = self.get_content_x();
+            let content_y = self.get_content_y();
+
+            let mut character_chunks = self.grid.read_changes(content_x, content_y);
+            for character_chunk in character_chunks.iter_mut() {
+                character_chunk.add_changed_colors(self.grid.changed_colors);
+                if self
+                    .grid
+                    .selection
+                    .contains_row(character_chunk.y.saturating_sub(content_y))
+                {
+                    let background_color = match self.colors.bg {
+                        PaletteColor::Rgb(rgb) => AnsiCode::RgbCode(rgb),
+                        PaletteColor::EightBit(col) => AnsiCode::ColorIndex(col),
+                    };
+                    character_chunk.add_selection_and_background(
+                        self.grid.selection,
+                        background_color,
+                        content_x,
+                        content_y,
+                    );
                 }
-                self.grid.clear_viewport_before_rendering = false;
             }
-            let max_width = self.get_content_columns();
-            for character_chunk in self.grid.read_changes() {
-                let pane_x = self.get_content_x();
-                let pane_y = self.get_content_y();
-                let chunk_absolute_x = pane_x + character_chunk.x;
-                let chunk_absolute_y = pane_y + character_chunk.y;
-                let terminal_characters = character_chunk.terminal_characters;
-                vte_output.push_str(&format!(
-                    "\u{1b}[{};{}H\u{1b}[m",
-                    chunk_absolute_y + 1,
-                    chunk_absolute_x + 1
-                )); // goto row/col and reset styles
-
-                let mut chunk_width = character_chunk.x;
-                for mut t_character in terminal_characters {
-                    // adjust the background of currently selected characters
-                    // doing it here is much easier than in grid
-                    if self.grid.selection.contains(character_chunk.y, chunk_width) {
-                        let color = match self.colors.bg {
-                            PaletteColor::Rgb(rgb) => AnsiCode::RgbCode(rgb),
-                            PaletteColor::EightBit(col) => AnsiCode::ColorIndex(col),
-                        };
-
-                        t_character.styles = t_character.styles.background(Some(color));
-                    }
-                    chunk_width += t_character.width;
-                    if chunk_width > max_width {
-                        break;
-                    }
-
-                    if let Some(new_styles) = character_styles
-                        .update_and_return_diff(&t_character.styles, self.grid.changed_colors)
-                    {
-                        vte_output.push_str(&new_styles.to_string());
-                    }
-                    vte_output.push(t_character.character);
-                }
-                character_styles.clear();
-            }
-            if let Some(last_frame) = &self.frame {
-                let frame = PaneFrame {
-                    geom: self.current_geom().into(),
-                    title: self
-                        .grid
-                        .title
-                        .clone()
-                        .unwrap_or_else(|| self.pane_title.clone()),
-                    scroll_position: self.grid.scrollback_position_and_length(),
-                    color: self.frame_color,
-                };
-                if &frame != last_frame {
-                    if !self.borderless {
-                        vte_output.push_str(&frame.render());
-                    }
-                    self.frame = Some(frame);
-                }
+            if self.grid.ring_bell {
+                let ring_bell = '\u{7}';
+                raw_vte_output.push(ring_bell);
+                self.grid.ring_bell = false;
             }
             self.set_should_render(false);
-            Some(vte_output)
+            Some((character_chunks, Some(raw_vte_output)))
         } else {
             None
         }
     }
+    fn render_frame(
+        &mut self,
+        client_id: ClientId,
+        frame_params: FrameParams,
+        input_mode: InputMode,
+    ) -> Option<(Vec<CharacterChunk>, Option<String>)> {
+        // TODO: remove the cursor stuff from here
+        let pane_title = if self.pane_name.is_empty()
+            && input_mode == InputMode::RenamePane
+            && frame_params.is_main_client
+        {
+            String::from("Enter name...")
+        } else if self.pane_name.is_empty() {
+            self.grid
+                .title
+                .clone()
+                .unwrap_or_else(|| self.pane_title.clone())
+        } else {
+            self.pane_name.clone()
+        };
+        let frame = PaneFrame::new(
+            self.current_geom().into(),
+            self.grid.scrollback_position_and_length(),
+            pane_title,
+            frame_params,
+        );
+        match self.frame.get(&client_id) {
+            // TODO: use and_then or something?
+            Some(last_frame) => {
+                if &frame != last_frame {
+                    if !self.borderless {
+                        let frame_output = frame.render();
+                        self.frame.insert(client_id, frame);
+                        Some(frame_output)
+                    } else {
+                        None
+                    }
+                } else {
+                    None
+                }
+            }
+            None => {
+                if !self.borderless {
+                    let frame_output = frame.render();
+                    self.frame.insert(client_id, frame);
+                    Some(frame_output)
+                } else {
+                    None
+                }
+            }
+        }
+    }
+    fn render_fake_cursor(
+        &mut self,
+        cursor_color: PaletteColor,
+        text_color: PaletteColor,
+    ) -> Option<String> {
+        let mut vte_output = None;
+        if let Some((cursor_x, cursor_y)) = self.cursor_coordinates() {
+            let mut character_under_cursor = self
+                .grid
+                .get_character_under_cursor()
+                .unwrap_or(EMPTY_TERMINAL_CHARACTER);
+            character_under_cursor.styles.background = Some(cursor_color.into());
+            character_under_cursor.styles.foreground = Some(text_color.into());
+            // we keep track of these so that we can clear them up later (see render function)
+            self.fake_cursor_locations.insert((cursor_y, cursor_x));
+            let mut fake_cursor = format!(
+                "\u{1b}[{};{}H\u{1b}[m{}",           // goto row column and clear styles
+                self.get_content_y() + cursor_y + 1, // + 1 because goto is 1 indexed
+                self.get_content_x() + cursor_x + 1,
+                &character_under_cursor.styles,
+            );
+            fake_cursor.push(character_under_cursor.character);
+            vte_output = Some(fake_cursor);
+        }
+        vte_output
+    }
+    fn render_terminal_title(&mut self, input_mode: InputMode) -> String {
+        let pane_title = if self.pane_name.is_empty() && input_mode == InputMode::RenamePane {
+            "Enter name..."
+        } else if self.pane_name.is_empty() {
+            self.grid.title.as_deref().unwrap_or(&self.pane_title)
+        } else {
+            &self.pane_name
+        };
+        make_terminal_title(pane_title)
+    }
+    fn update_name(&mut self, name: &str) {
+        match name {
+            "\0" => {
+                self.pane_name = String::new();
+            }
+            "\u{007F}" | "\u{0008}" => {
+                //delete and backspace keys
+                self.pane_name.pop();
+            }
+            c => {
+                self.pane_name.push_str(c);
+            }
+        }
+    }
     fn pid(&self) -> PaneId {
         PaneId::Terminal(self.pid)
     }
@@ -308,11 +384,11 @@ impl Pane for TerminalPane {
         self.geom.y -= count;
         self.reflow_lines();
     }
-    fn scroll_up(&mut self, count: usize) {
+    fn scroll_up(&mut self, count: usize, _client_id: ClientId) {
         self.grid.move_viewport_up(count);
         self.set_should_render(true);
     }
-    fn scroll_down(&mut self, count: usize) {
+    fn scroll_down(&mut self, count: usize, _client_id: ClientId) {
         self.grid.move_viewport_down(count);
         self.set_should_render(true);
     }
@@ -346,12 +422,12 @@ impl Pane for TerminalPane {
         self.grid.pending_messages_to_pty.drain(..).collect()
     }
 
-    fn start_selection(&mut self, start: &Position) {
+    fn start_selection(&mut self, start: &Position, _client_id: ClientId) {
         self.grid.start_selection(start);
         self.set_should_render(true);
     }
 
-    fn update_selection(&mut self, to: &Position) {
+    fn update_selection(&mut self, to: &Position, _client_id: ClientId) {
         let should_scroll = self.selection_scrolled_at.elapsed()
             >= time::Duration::from_millis(SELECTION_SCROLL_INTERVAL_MS);
         // TODO: check how far up/down mouse is relative to pane, to increase scroll lines?
@@ -368,7 +444,7 @@ impl Pane for TerminalPane {
         self.set_should_render(true);
     }
 
-    fn end_selection(&mut self, end: Option<&Position>) {
+    fn end_selection(&mut self, end: &Position, _client_id: ClientId) {
         self.grid.end_selection(end);
         self.set_should_render(true);
     }
@@ -381,12 +457,8 @@ impl Pane for TerminalPane {
         self.grid.get_selected_text()
     }
 
-    fn set_frame(&mut self, frame: bool) {
-        self.frame = if frame {
-            Some(PaneFrame::default())
-        } else {
-            None
-        };
+    fn set_frame(&mut self, _frame: bool) {
+        self.frame.clear();
     }
 
     fn set_content_offset(&mut self, offset: Offset) {
@@ -394,16 +466,16 @@ impl Pane for TerminalPane {
         self.reflow_lines();
     }
 
-    fn set_boundary_color(&mut self, color: Option<PaletteColor>) {
-        self.frame_color = color;
-        self.set_should_render(true);
-    }
     fn set_borderless(&mut self, borderless: bool) {
         self.borderless = borderless;
     }
     fn borderless(&self) -> bool {
         self.borderless
     }
+
+    fn mouse_mode(&self) -> bool {
+        self.grid.mouse_mode
+    }
 }
 
 impl TerminalPane {
@@ -412,16 +484,18 @@ impl TerminalPane {
         position_and_size: PaneGeom,
         palette: Palette,
         pane_index: usize,
+        pane_name: String,
+        link_handler: Rc<RefCell<LinkHandler>>,
     ) -> TerminalPane {
         let initial_pane_title = format!("Pane #{}", pane_index);
         let grid = Grid::new(
             position_and_size.rows.as_usize(),
             position_and_size.cols.as_usize(),
             palette,
+            link_handler,
         );
         TerminalPane {
-            frame: None,
-            frame_color: None,
+            frame: HashMap::new(),
             content_offset: Offset::default(),
             pid,
             grid,
@@ -433,7 +507,9 @@ impl TerminalPane {
             colors: palette,
             selection_scrolled_at: time::Instant::now(),
             pane_title: initial_pane_title,
+            pane_name,
             borderless: false,
+            fake_cursor_locations: HashSet::new(),
         }
     }
     pub fn get_x(&self) -> usize {
